﻿using System;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace HeuristicLab.Problems.DataAnalysis.Tests {
  [TestClass]
  public class AlgebraicIntervalTest {
    private readonly AlgebraicInterval a = new AlgebraicInterval(-1, 1);
    private readonly AlgebraicInterval b = new AlgebraicInterval(-2, 2);
    private readonly AlgebraicInterval c = new AlgebraicInterval(0, 3);
    private readonly AlgebraicInterval d = new AlgebraicInterval(1, 3);
    private readonly AlgebraicInterval e = new AlgebraicInterval(4, 6);

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void AlgebraicIntervalAdd() {
      //add        [x1,x2] + [y1,y2] = [x1 + y1,x2 + y2]

      // [-1,1] + [-2,2] = [-3,3]
      AssertAreEqualInterval(Add(a, b), new AlgebraicInterval(-3, 3));
      //([-1, 1] + [-2, 2]) + [0, 3] = [-3, 6]
      AssertAreEqualInterval(Add(Add(a, b), c), new AlgebraicInterval(-3, 6));
      //([-1, 1] + [0, 3]) + [-2, 2] = [-3, 6]
      AssertAreEqualInterval(Add(Add(a, c), b), new AlgebraicInterval(-3, 6));
    }

    private void AssertAreEqualInterval(AlgebraicInterval a, AlgebraicInterval b) {
      if (double.IsNaN(a.LowerBound.Value)) {
        Assert.IsTrue(double.IsNaN(b.LowerBound.Value));
      } else {
        Assert.AreEqual(a.LowerBound.Value.Value, b.LowerBound.Value.Value, Math.Abs(a.LowerBound.Value.Value) * 1e-4); // relative error < 0.1%
      }

      if (double.IsNaN(a.UpperBound.Value)) {
        Assert.IsTrue(double.IsNaN(b.UpperBound.Value));
      } else {
        Assert.AreEqual(a.UpperBound.Value.Value, b.UpperBound.Value.Value, Math.Abs(a.UpperBound.Value.Value) * 1e-4);
      }
    }

    private static AlgebraicInterval Add(AlgebraicInterval a, AlgebraicInterval b) {
      return a.Clone().Add(b);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void AlgebraicIntervalSub() {
      //subtract   [x1,x2] − [y1,y2] = [x1 − y2,x2 − y1]

      //[-1, 1] - [-2, 2] = [-3, 3]
      AssertAreEqualInterval(Subtract(a, b), new AlgebraicInterval(-3, 3));
      //([-1, 1] - [-2, 2]) - [0, 3] = [-6, 3]
      AssertAreEqualInterval(Subtract(Subtract(a, b), c), new AlgebraicInterval(-6, 3));
      //([-1, 1] - [0, 3]) - [-2, 2] = [-6, 3]
      AssertAreEqualInterval(Subtract(Subtract(a, c), b), new AlgebraicInterval(-6, 3));
    }

    private AlgebraicInterval Subtract(AlgebraicInterval a, AlgebraicInterval b) {
      return a.Clone().Sub(b);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void AlgebraicIntervalMutlipy() {
      //multiply   [x1,x2] * [y1,y2] = [min(x1*y1,x1*y2,x2*y1,x2*y2),max(x1*y1,x1*y2,x2*y1,x2*y2)]

      //[-1, 1] * [-2, 2] = [-2, 2]
      AssertAreEqualInterval(Multiply(a, b), new AlgebraicInterval(-2, 2));
      //([-1, 1] * [-2, 2]) * [0, 3] = [-6, 6]
      AssertAreEqualInterval(Multiply(Multiply(a, b), c), new AlgebraicInterval(-6, 6));
      //([-1, 1] * [0, 3]) * [-2, 2] = [-6, 6]
      AssertAreEqualInterval(Multiply(Multiply(a, c), b), new AlgebraicInterval(-6, 6));

      // [-2, 0] * [-2, 0]  = [0, 4]
      AssertAreEqualInterval(new AlgebraicInterval(0, 4), Multiply(new AlgebraicInterval(-2, 0), new AlgebraicInterval(-2, 0)));
    }

    private AlgebraicInterval Multiply(AlgebraicInterval a, AlgebraicInterval b) {
      return a.Clone().Mul(b);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void AlgebraicIntervalDivide() {
      //divide  [x1, x2] / [y1, y2] = [x1, x2] * (1/[y1, y2]), where 1 / [y1,y2] = [1 / y2,1 / y1] if 0 not in [y_1, y_2].

      //[4, 6] / [1, 3] = [4/3, 6]
      AssertAreEqualInterval(Divide(e, d), new AlgebraicInterval(4.0 / 3.0, 6));
      //([4, 6] / [1, 3]) / [1, 3] = [4/9, 6]
      AssertAreEqualInterval(Divide(Divide(e, d), d), new AlgebraicInterval(4.0 / 9.0, 6));
      //[4, 6] / [0, 3] = [4/3, +Inf]
      AssertAreEqualInterval(Divide(e, c), new AlgebraicInterval(4.0 / 3.0, double.PositiveInfinity));
      //[-1, 1] / [0, 3] = [+Inf, -Inf]
      AssertAreEqualInterval(Divide(a, c), new AlgebraicInterval(double.NegativeInfinity, double.PositiveInfinity));

      //Division by 0 ==> IsInfiniteOrUndefined == true
      AssertAreEqualInterval(Divide(d, b), new AlgebraicInterval(double.NegativeInfinity, double.PositiveInfinity));

      // [0, 1] / [0, 1]
      AssertAreEqualInterval(new AlgebraicInterval(0, 1).Div(new AlgebraicInterval(0, 1)), new AlgebraicInterval(1, double.PositiveInfinity)); // returns [NaN, PositiveInfinity]

    }

    private AlgebraicInterval Divide(AlgebraicInterval e, AlgebraicInterval d) {
      return e.Clone().Div(d);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void AlgebraicIntervalSine() {
      //sine depends on interval
      //sin([0, 2*pi]) = [-1, 1]
      AssertAreEqualInterval(Sine(new AlgebraicInterval(0, 2 * Math.PI)), new AlgebraicInterval(-1, 1));
      //sin([-pi/2, pi/2]) = [sin(-pi/2), sin(pi/2)]
      AssertAreEqualInterval(Sine(new AlgebraicInterval(-1 * Math.PI / 2, Math.PI / 2)), new AlgebraicInterval(-1, 1));
      //sin([0, pi/2]) = [sin(0), sin(pi/2)]
      AssertAreEqualInterval(Sine(new AlgebraicInterval(0, Math.PI / 2)), new AlgebraicInterval(0, 1));
      //sin([pi, 3*pi/2]) = [sin(pi), sin(3*pi/2)]
      AssertAreEqualInterval(Sine(new AlgebraicInterval(Math.PI, 3 * Math.PI / 2)), new AlgebraicInterval(-1, 0));
      AssertAreEqualInterval(Sine(new AlgebraicInterval(1, 2)), new AlgebraicInterval(Math.Min(Math.Sin(1), Math.Sin(2)), 1));
      AssertAreEqualInterval(Sine(new AlgebraicInterval(1, 3)), new AlgebraicInterval(Math.Min(Math.Sin(1), Math.Sin(3)), 1));
      AssertAreEqualInterval(Sine(new AlgebraicInterval(Math.PI, 5 * Math.PI / 2)), new AlgebraicInterval(-1, 1));
    }

    private AlgebraicInterval Sine(AlgebraicInterval algebraicInterval) {
      return algebraicInterval.Sin();
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void AlgebraicIntervalCosine() {
      //Cosine uses sine Sine(Subtract(a, new AlgebraicInterval(Math.PI / 2, Math.PI / 2)));
      AssertAreEqualInterval(Cosine(new AlgebraicInterval(0, 2 * Math.PI)), new AlgebraicInterval(-1, 1));
      AssertAreEqualInterval(new AlgebraicInterval(-1, 1), Cosine(new AlgebraicInterval(Math.PI, 4 * Math.PI / 2)));

      // some random expression that were problematic
      AssertAreEqualInterval(new AlgebraicInterval(-0.0326026961803847, 1), new AlgebraicInterval(1, 2).Scale(1.73269159006613).Add(9.23027422279119).Cos());
      AssertAreEqualInterval(new AlgebraicInterval(0.54030230586814, 1), new AlgebraicInterval(1, 2).Scale(3.39318455754522).Cos().IntPower(2).Cos());
      AssertAreEqualInterval(new AlgebraicInterval(-0.825841274068331, 1), new AlgebraicInterval(1, 2).Scale(-16.7140105366707 * -0.223806041028346).Cos());
      AssertAreEqualInterval(new AlgebraicInterval(0, 1), new AlgebraicInterval(1, 2).Scale(1.03513653909418).Exp().Cos().IntPower(2));
      AssertAreEqualInterval(new AlgebraicInterval(-1, 1), new AlgebraicInterval(1, 2).Scale(0.535386838272755).Exp().IntPower(2).Cos());
    }

    private AlgebraicInterval Cosine(AlgebraicInterval algebraicInterval) {
      return algebraicInterval.Cos();
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void AlgebraicIntervalLog() {
      //Log([3, 5]) = [log(3), log(5)]
      AssertAreEqualInterval(new AlgebraicInterval(Math.Log(3), Math.Log(5)), Logarithm(new AlgebraicInterval(3, 5)));
      //Log([0.5, 1]) = [log(0.5), log(1)]
      AssertAreEqualInterval(new AlgebraicInterval(Math.Log(0.5), 0), Logarithm(new AlgebraicInterval(0.5, 1)));
      //Log([-1, 5]) = [NaN, log(5)]
      var result = Logarithm(new AlgebraicInterval(-1, 5));
      AssertAreEqualInterval(new AlgebraicInterval(double.NaN, Math.Log(5)), result);

    }

    private AlgebraicInterval Logarithm(AlgebraicInterval algebraicInterval) {
      return algebraicInterval.Log();
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void AlgebraicIntervalExp() {
      //Exp([0, 1]) = [exp(0), exp(1)]
      AssertAreEqualInterval(new AlgebraicInterval(1, Math.Exp(1)), Exponential(new AlgebraicInterval(0, 1)));
    }

    private AlgebraicInterval Exponential(AlgebraicInterval algebraicInterval) {
      return algebraicInterval.Exp();
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void AlgebraicIntervalSqr() {
      AssertAreEqualInterval(new AlgebraicInterval(1, 4), new AlgebraicInterval(1, 2).IntPower(2));
      AssertAreEqualInterval(new AlgebraicInterval(1, 4), new AlgebraicInterval(-2, -1).IntPower(2));
      AssertAreEqualInterval(new AlgebraicInterval(0, 4), new AlgebraicInterval(-2, 2).IntPower(2));

      AssertAreEqualInterval(new AlgebraicInterval(0, 16), new AlgebraicInterval(-2, 2).Mul(new AlgebraicInterval(2, 2)).IntPower(2));
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void AlgebraicIntervalSqrt() {
      AssertAreEqualInterval(new AlgebraicInterval(1, 2), new AlgebraicInterval(1, 4).IntRoot(2));
      AssertAreEqualInterval(new AlgebraicInterval(double.NaN, double.NaN), new AlgebraicInterval(-4, -1).IntRoot(2));
      AssertAreEqualInterval(new AlgebraicInterval(0, 1), new AlgebraicInterval(0, 1).IntRoot(2));
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void AlgebraicIntervalCube() {
      AssertAreEqualInterval(new AlgebraicInterval(1, 8), new AlgebraicInterval(1, 2).IntPower(3));
      AssertAreEqualInterval(new AlgebraicInterval(-8, -1), new AlgebraicInterval(-2, -1).IntPower(3));
      AssertAreEqualInterval(new AlgebraicInterval(-8, 8), new AlgebraicInterval(-2, 2).IntPower(3));
    }


    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void AlgebraicIntervalAbs() {
      AssertAreEqualInterval(new AlgebraicInterval(1, 2), new AlgebraicInterval(1, 2).Abs());
      AssertAreEqualInterval(new AlgebraicInterval(1, 2), new AlgebraicInterval(-2, -1).Abs());
      AssertAreEqualInterval(new AlgebraicInterval(0, 2), new AlgebraicInterval(-2, 2).Abs());
      AssertAreEqualInterval(new AlgebraicInterval(0, double.PositiveInfinity), new AlgebraicInterval(double.NegativeInfinity, double.PositiveInfinity).Abs());
      AssertAreEqualInterval(new AlgebraicInterval(double.NaN, 1), new AlgebraicInterval(-1, double.NaN).Abs()); // XXX unclear how we should handle NaN
      AssertAreEqualInterval(new AlgebraicInterval(double.NaN, 1), new AlgebraicInterval(double.NaN, -1).Abs());
    }


    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void AlgebraicIntervalCbrt() {
      AssertAreEqualInterval(new AlgebraicInterval(1, 2), new AlgebraicInterval(1, 8).IntRoot(3));
      AssertAreEqualInterval(new AlgebraicInterval(-2, -1), new AlgebraicInterval(-8, -1).IntRoot(3));
      AssertAreEqualInterval(new AlgebraicInterval(0, 1), new AlgebraicInterval(0, 1).IntRoot(3));
    }
  }
}
